@@ -68,8 +68,9 @@
 #define MAX_THREADS 500
 #define CLUSTER_SLOTS 16384
 #define CONFIG_LATENCY_HISTOGRAM_MIN_VALUE 10L          /* >= 10 usecs */
-#define CONFIG_LATENCY_HISTOGRAM_MAX_VALUE 3000000L          /* <= 30 secs(us precision) */
+#define CONFIG_LATENCY_HISTOGRAM_MAX_VALUE 3000000L          /* <= 3 secs(us precision) */
 #define CONFIG_LATENCY_HISTOGRAM_INSTANT_MAX_VALUE 3000000L   /* <= 3 secs(us precision) */
+#define SHOW_THROUGHPUT_INTERVAL 250  /* 250ms */
 
 #define CLIENT_GET_EVENTLOOP(c) \
     (c->thread_id >= 0 ? config.threads[c->thread_id]->el : config.el)
@@ -80,8 +81,7 @@ struct redisConfig;
 
 static struct config {
     aeEventLoop *el;
-    const char *hostip;
-    int hostport;
+    cliConnInfo conn_info;
     const char *hostsocket;
     int tls;
     struct cliSSLconfig sslconfig;
@@ -99,21 +99,17 @@ static struct config {
     int randomkeys_keyspacelen;
     int keepalive;
     int pipeline;
-    int showerrors;
     long long start;
     long long totlatency;
-    long long *latency;
     const char *title;
     list *clients;
     int quiet;
     int csv;
     int loop;
     int idlemode;
-    int dbnum;
-    sds dbnumstr;
+    sds input_dbnumstr;
     char *tests;
-    char *auth;
-    const char *user;
+    int stdinarg; /* get last arg from stdin. (-x option) */
     int precision;
     int num_threads;
     struct benchmarkThread **threads;
@@ -129,6 +125,7 @@ static struct config {
     int enable_tracking;
     pthread_mutex_t liveclients_mutex;
     pthread_mutex_t is_updating_slots_mutex;
+    int resp3; /* use RESP3 */
 } config;
 
 typedef struct _client {
@@ -224,8 +221,7 @@ static sds benchmarkVersion(void) {
 
 /* Dict callbacks */
 static uint64_t dictSdsHash(const void *key);
-static int dictSdsKeyCompare(void *privdata, const void *key1,
-    const void *key2);
+static int dictSdsKeyCompare(dict *d, const void *key1, const void *key2);
 
 /* Implementation */
 static long long ustime(void) {
@@ -252,11 +248,10 @@ static uint64_t dictSdsHash(const void *key) {
     return dictGenHashFunction((unsigned char*)key, sdslen((char*)key));
 }
 
-static int dictSdsKeyCompare(void *privdata, const void *key1,
-        const void *key2)
+static int dictSdsKeyCompare(dict *d, const void *key1, const void *key2)
 {
     int l1,l2;
-    DICT_NOTUSED(privdata);
+    UNUSED(d);
 
     l1 = sdslen((sds)key1);
     l2 = sdslen((sds)key2);
@@ -264,13 +259,6 @@ static int dictSdsKeyCompare(void *privdata, const void *key1,
     return memcmp(key1, key2, l1) == 0;
 }
 
-/* _serverAssert is needed by dict */
-void _serverAssert(const char *estr, const char *file, int line) {
-    fprintf(stderr, "=== ASSERTION FAILED ===");
-    fprintf(stderr, "==> %s:%d '%s' is not true",file,line,estr);
-    *((char*)-1) = 'x';
-}
-
 static redisContext *getRedisContext(const char *ip, int port,
                                      const char *hostsocket)
 {
@@ -296,19 +284,21 @@ static redisContext *getRedisContext(const char *ip, int port,
             goto cleanup;
         }
     }
-    if (config.auth == NULL)
+    if (config.conn_info.auth == NULL)
         return ctx;
-    if (config.user == NULL)
-        reply = redisCommand(ctx,"AUTH %s", config.auth);
+    if (config.conn_info.user == NULL)
+        reply = redisCommand(ctx,"AUTH %s", config.conn_info.auth);
     else
-        reply = redisCommand(ctx,"AUTH %s %s", config.user, config.auth);
+        reply = redisCommand(ctx,"AUTH %s %s", config.conn_info.user, config.conn_info.auth);
     if (reply != NULL) {
         if (reply->type == REDIS_REPLY_ERROR) {
             if (hostsocket == NULL)
                 fprintf(stderr, "Node %s:%d replied with error:\n%s\n", ip, port, reply->str);
             else
                 fprintf(stderr, "Node %s replied with error:\n%s\n", hostsocket, reply->str);
-            goto cleanup;
+            freeReplyObject(reply);
+            redisFree(ctx);
+            exit(1);
         }
         freeReplyObject(reply);
         return ctx;
@@ -367,9 +357,16 @@ fail:
     fprintf(stderr, "ERROR: failed to fetch CONFIG from ");
     if (hostsocket == NULL) fprintf(stderr, "%s:%d\n", ip, port);
     else fprintf(stderr, "%s\n", hostsocket);
+    int abort_test = 0;
+    if (reply && reply->type == REDIS_REPLY_ERROR &&
+        (!strncmp(reply->str,"NOAUTH",6) ||
+         !strncmp(reply->str,"WRONGPASS",9) ||
+         !strncmp(reply->str,"NOPERM",6)))
+        abort_test = 1;
     freeReplyObject(reply);
     redisFree(c);
     freeRedisConfig(cfg);
+    if (abort_test) exit(1);
     return NULL;
 }
 static void freeRedisConfig(redisConfig *cfg) {
@@ -514,44 +511,39 @@ static void readHandler(aeEventLoop *el, int fd, void *privdata, int mask) {
                     exit(1);
                 }
                 redisReply *r = reply;
-                int is_err = (r->type == REDIS_REPLY_ERROR);
-
-                if (is_err && config.showerrors) {
-                    /* TODO: static lasterr_time not thread-safe */
-                    static time_t lasterr_time = 0;
-                    time_t now = time(NULL);
-                    if (lasterr_time != now) {
-                        lasterr_time = now;
+                if (r->type == REDIS_REPLY_ERROR) {
+                    /* Try to update slots configuration if reply error is
+                    * MOVED/ASK/CLUSTERDOWN and the key(s) used by the command
+                    * contain(s) the slot hash tag.
+                    * If the error is not topology-update related then we
+                    * immediately exit to avoid false results. */
+                    if (c->cluster_node && c->staglen) {
+                        int fetch_slots = 0, do_wait = 0;
+                        if (!strncmp(r->str,"MOVED",5) || !strncmp(r->str,"ASK",3))
+                            fetch_slots = 1;
+                        else if (!strncmp(r->str,"CLUSTERDOWN",11)) {
+                            /* Usually the cluster is able to recover itself after
+                            * a CLUSTERDOWN error, so try to sleep one second
+                            * before requesting the new configuration. */
+                            fetch_slots = 1;
+                            do_wait = 1;
+                            fprintf(stderr, "Error from server %s:%d: %s.\n",
+                                    c->cluster_node->ip,
+                                    c->cluster_node->port,
+                                    r->str);
+                        }
+                        if (do_wait) sleep(1);
+                        if (fetch_slots && !fetchClusterSlotsConfiguration(c))
+                            exit(1);
+                    } else {
                         if (c->cluster_node) {
-                            printf("Error from server %s:%d: %s\n",
-                                   c->cluster_node->ip,
-                                   c->cluster_node->port,
-                                   r->str);
-                        } else printf("Error from server: %s\n", r->str);
-                    }
-                }
-
-                /* Try to update slots configuration if reply error is
-                 * MOVED/ASK/CLUSTERDOWN and the key(s) used by the command
-                 * contain(s) the slot hash tag. */
-                if (is_err && c->cluster_node && c->staglen) {
-                    int fetch_slots = 0, do_wait = 0;
-                    if (!strncmp(r->str,"MOVED",5) || !strncmp(r->str,"ASK",3))
-                        fetch_slots = 1;
-                    else if (!strncmp(r->str,"CLUSTERDOWN",11)) {
-                        /* Usually the cluster is able to recover itself after
-                         * a CLUSTERDOWN error, so try to sleep one second
-                         * before requesting the new configuration. */
-                        fetch_slots = 1;
-                        do_wait = 1;
-                        printf("Error from server %s:%d: %s\n",
-                               c->cluster_node->ip,
-                               c->cluster_node->port,
-                               r->str);
-                    }
-                    if (do_wait) sleep(1);
-                    if (fetch_slots && !fetchClusterSlotsConfiguration(c))
+                            fprintf(stderr, "Error from server %s:%d: %s\n",
+                                 c->cluster_node->ip,
+                                 c->cluster_node->port,
+                                 r->str);
+                        } else fprintf(stderr, "Error from server: %s\n", r->str);
                         exit(1);
+                    }
                 }
 
                 freeReplyObject(reply);
@@ -641,6 +633,9 @@ static void writeHandler(aeEventLoop *el, int fd, void *privdata, int mask) {
                         fprintf(stderr, "Error writing to the server: %s\n", strerror(errno));
                     freeClient(c);
                     return;
+                } else if (nwritten > 0) {
+                    c->written += nwritten;
+                    return;
                 }
             } else {
                 aeDeleteFileEvent(el,c->context->fd,AE_WRITABLE);
@@ -682,8 +677,8 @@ static client createClient(char *cmd, size_t len, client from, int thread_id) {
     c->cluster_node = NULL;
     if (config.hostsocket == NULL || is_cluster_client) {
         if (!is_cluster_client) {
-            ip = config.hostip;
-            port = config.hostport;
+            ip = config.conn_info.hostip;
+            port = config.conn_info.hostport;
         } else {
             int node_idx = 0;
             if (config.num_threads < config.cluster_node_count)
@@ -727,14 +722,14 @@ static client createClient(char *cmd, size_t len, client from, int thread_id) {
      * These commands are discarded after the first response, so if the client is
      * reused the commands will not be used again. */
     c->prefix_pending = 0;
-    if (config.auth) {
+    if (config.conn_info.auth) {
         char *buf = NULL;
         int len;
-        if (config.user == NULL)
-            len = redisFormatCommand(&buf, "AUTH %s", config.auth);
+        if (config.conn_info.user == NULL)
+            len = redisFormatCommand(&buf, "AUTH %s", config.conn_info.auth);
         else
             len = redisFormatCommand(&buf, "AUTH %s %s",
-                                     config.user, config.auth);
+                                     config.conn_info.user, config.conn_info.auth);
         c->obuf = sdscatlen(c->obuf, buf, len);
         free(buf);
         c->prefix_pending++;
@@ -752,11 +747,20 @@ static client createClient(char *cmd, size_t len, client from, int thread_id) {
      * buffer with the SELECT command, that will be discarded the first
      * time the replies are received, so if the client is reused the
      * SELECT command will not be used again. */
-    if (config.dbnum != 0 && !is_cluster_client) {
+    if (config.conn_info.input_dbnum != 0 && !is_cluster_client) {
         c->obuf = sdscatprintf(c->obuf,"*2\r\n$6\r\nSELECT\r\n$%d\r\n%s\r\n",
-            (int)sdslen(config.dbnumstr),config.dbnumstr);
+            (int)sdslen(config.input_dbnumstr),config.input_dbnumstr);
         c->prefix_pending++;
     }
+
+    if (config.resp3) {
+        char *buf = NULL;
+        int len = redisFormatCommand(&buf, "HELLO 3");
+        c->obuf = sdscatlen(c->obuf, buf, len);
+        free(buf);
+        c->prefix_pending++;
+    }
+
     c->prefixlen = sdslen(c->obuf);
     /* Append the request itself. */
     if (from) {
@@ -842,6 +846,10 @@ static client createClient(char *cmd, size_t len, client from, int thread_id) {
     }
     if (config.idlemode == 0)
         aeCreateFileEvent(el,c->context->fd,AE_WRITABLE,writeHandler,c);
+    else
+        /* In idle mode, clients still need to register readHandler for catching errors */
+        aeCreateFileEvent(el,c->context->fd,AE_READABLE,readHandler,c);
+
     listAddNodeTail(config.clients,c);
     atomicIncr(config.liveclients, 1);
     atomicGet(config.slots_last_update, c->slots_last_update);
@@ -980,7 +988,7 @@ static void startBenchmarkThreads() {
         pthread_join(config.threads[i]->thread, NULL);
 }
 
-static void benchmark(char *title, char *cmd, int len) {
+static void benchmark(const char *title, char *cmd, int len) {
     client c;
 
     config.title = title;
@@ -1025,7 +1033,7 @@ static benchmarkThread *createBenchmarkThread(int index) {
     if (thread == NULL) return NULL;
     thread->index = index;
     thread->el = aeCreateEventLoop(1024*10);
-    aeCreateTimeEvent(thread->el,1,showThroughput,NULL,NULL);
+    aeCreateTimeEvent(thread->el,1,showThroughput,(void *)thread,NULL);
     return thread;
 }
 
@@ -1087,9 +1095,9 @@ static void freeClusterNode(clusterNode *node) {
         zfree(node->importing);
     }
     /* If the node is not the reference node, that uses the address from
-     * config.hostip and config.hostport, then the node ip has been
+     * config.conn_info.hostip and config.conn_info.hostport, then the node ip has been
      * allocated by fetchClusterConfiguration, so it must be freed. */
-    if (node->ip && strcmp(node->ip, config.hostip) != 0) sdsfree(node->ip);
+    if (node->ip && strcmp(node->ip, config.conn_info.hostip) != 0) sdsfree(node->ip);
     if (node->redis_config != NULL) freeRedisConfig(node->redis_config);
     zfree(node->slots);
     zfree(node);
@@ -1114,16 +1122,19 @@ static clusterNode **addClusterNode(clusterNode *node) {
     return config.cluster_nodes;
 }
 
+/* TODO: This should be refactored to use CLUSTER SLOTS, the migrating/importing
+ * information is anyway not used.
+ */
 static int fetchClusterConfiguration() {
     int success = 1;
     redisContext *ctx = NULL;
     redisReply *reply =  NULL;
-    ctx = getRedisContext(config.hostip, config.hostport, config.hostsocket);
+    ctx = getRedisContext(config.conn_info.hostip, config.conn_info.hostport, config.hostsocket);
     if (ctx == NULL) {
         exit(1);
     }
-    clusterNode *firstNode = createClusterNode((char *) config.hostip,
-                                               config.hostport);
+    clusterNode *firstNode = createClusterNode((char *) config.conn_info.hostip,
+                                               config.conn_info.hostport);
     if (!firstNode) {success = 0; goto cleanup;}
     reply = redisCommand(ctx, "CLUSTER NODES");
     success = (reply != NULL);
@@ -1132,7 +1143,7 @@ static int fetchClusterConfiguration() {
     if (!success) {
         if (config.hostsocket == NULL) {
             fprintf(stderr, "Cluster node %s:%d replied with error:\n%s\n",
-                    config.hostip, config.hostport, reply->str);
+                    config.conn_info.hostip, config.conn_info.hostport, reply->str);
         } else {
             fprintf(stderr, "Cluster node %s replied with error:\n%s\n",
                     config.hostsocket, reply->str);
@@ -1175,7 +1186,7 @@ static int fetchClusterConfiguration() {
         clusterNode *node = NULL;
         char *ip = NULL;
         int port = 0;
-        char *paddr = strchr(addr, ':');
+        char *paddr = strrchr(addr, ':');
         if (paddr != NULL) {
             *paddr = '\0';
             ip = addr;
@@ -1261,8 +1272,9 @@ static int fetchClusterConfiguration() {
             }
         }
         if (node->slots_count == 0) {
-            printf("WARNING: master node %s:%d has no slots, skipping...\n",
-                   node->ip, node->port);
+            fprintf(stderr,
+                    "WARNING: Master node %s:%d has no slots, skipping...\n",
+                    node->ip, node->port);
             continue;
         }
         if (!addClusterNode(node)) {
@@ -1294,8 +1306,8 @@ static int fetchClusterSlotsConfiguration(client c) {
     atomicGetIncr(config.is_fetching_slots, is_fetching_slots, 1);
     if (is_fetching_slots) return -1; //TODO: use other codes || errno ?
     atomicSet(config.is_fetching_slots, 1);
-    if (config.showerrors)
-        printf("Cluster slots configuration changed, fetching new one...\n");
+    fprintf(stderr,
+            "WARNING: Cluster slots configuration changed, fetching new one...\n");
     const char *errmsg = "Failed to update cluster slots configuration";
     static dictType dtype = {
         dictSdsHash,               /* hash function */
@@ -1307,7 +1319,7 @@ static int fetchClusterSlotsConfiguration(client c) {
         NULL                       /* allow to expand */
     };
     /* printf("[%d] fetchClusterSlotsConfiguration\n", c->thread_id); */
-    dict *masters = dictCreate(&dtype, NULL);
+    dict *masters = dictCreate(&dtype);
     redisContext *ctx = NULL;
     for (i = 0; i < (size_t) config.cluster_node_count; i++) {
         clusterNode *node = config.cluster_nodes[i];
@@ -1405,7 +1417,7 @@ static void genBenchmarkRandomData(char *data, int count) {
 }
 
 /* Returns number of consumed options. */
-int parseOptions(int argc, const char **argv) {
+int parseOptions(int argc, char **argv) {
     int i;
     int lastarg;
     int exit_status = 1;
@@ -1429,19 +1441,27 @@ int parseOptions(int argc, const char **argv) {
             config.keepalive = atoi(argv[++i]);
         } else if (!strcmp(argv[i],"-h")) {
             if (lastarg) goto invalid;
-            config.hostip = strdup(argv[++i]);
+            sdsfree(config.conn_info.hostip);
+            config.conn_info.hostip = sdsnew(argv[++i]);
         } else if (!strcmp(argv[i],"-p")) {
             if (lastarg) goto invalid;
-            config.hostport = atoi(argv[++i]);
+            config.conn_info.hostport = atoi(argv[++i]);
         } else if (!strcmp(argv[i],"-s")) {
             if (lastarg) goto invalid;
             config.hostsocket = strdup(argv[++i]);
+        } else if (!strcmp(argv[i],"-x")) {
+            config.stdinarg = 1;
         } else if (!strcmp(argv[i],"-a") ) {
             if (lastarg) goto invalid;
-            config.auth = strdup(argv[++i]);
+            config.conn_info.auth = sdsnew(argv[++i]);
         } else if (!strcmp(argv[i],"--user")) {
             if (lastarg) goto invalid;
-            config.user = argv[++i];
+            config.conn_info.user = sdsnew(argv[++i]);
+        } else if (!strcmp(argv[i],"-u") && !lastarg) {
+            parseRedisUri(argv[++i],"redis-benchmark",&config.conn_info,&config.tls);
+            config.input_dbnumstr = sdsfromlonglong(config.conn_info.input_dbnum);
+        } else if (!strcmp(argv[i],"-3")) {
+            config.resp3 = 1;
         } else if (!strcmp(argv[i],"-d")) {
             if (lastarg) goto invalid;
             config.datasize = atoi(argv[++i]);
@@ -1471,7 +1491,9 @@ int parseOptions(int argc, const char **argv) {
         } else if (!strcmp(argv[i],"-I")) {
             config.idlemode = 1;
         } else if (!strcmp(argv[i],"-e")) {
-            config.showerrors = 1;
+            fprintf(stderr,
+                    "WARNING: -e option has no effect. "
+                    "We now immediately exit on error to avoid false results.\n");
         } else if (!strcmp(argv[i],"-t")) {
             if (lastarg) goto invalid;
             /* We get the list of tests to run as a string in the form
@@ -1485,8 +1507,8 @@ int parseOptions(int argc, const char **argv) {
             sdstolower(config.tests);
         } else if (!strcmp(argv[i],"--dbnum")) {
             if (lastarg) goto invalid;
-            config.dbnum = atoi(argv[++i]);
-            config.dbnumstr = sdsfromlonglong(config.dbnum);
+            config.conn_info.input_dbnum = atoi(argv[++i]);
+            config.input_dbnumstr = sdsfromlonglong(config.conn_info.input_dbnum);
         } else if (!strcmp(argv[i],"--precision")) {
             if (lastarg) goto invalid;
             config.precision = atoi(argv[++i]);
@@ -1496,8 +1518,9 @@ int parseOptions(int argc, const char **argv) {
              if (lastarg) goto invalid;
              config.num_threads = atoi(argv[++i]);
              if (config.num_threads > MAX_THREADS) {
-                printf("WARNING: too many threads, limiting threads to %d.\n",
-                       MAX_THREADS);
+                 fprintf(stderr,
+                         "WARNING: Too many threads, limiting threads to %d.\n",
+                         MAX_THREADS);
                 config.num_threads = MAX_THREADS;
              } else if (config.num_threads < 0) config.num_threads = 0;
         } else if (!strcmp(argv[i],"--cluster")) {
@@ -1552,37 +1575,48 @@ invalid:
 
 usage:
     printf(
-"Usage: redis-benchmark [-h <host>] [-p <port>] [-c <clients>] [-n <requests>] [-k <boolean>]\n\n"
+"%s%s", /* Split to avoid strings longer than 4095 (-Woverlength-strings). */
+"Usage: redis-benchmark [OPTIONS] [COMMAND ARGS...]\n\n"
+"Options:\n"
 " -h <hostname>      Server hostname (default 127.0.0.1)\n"
 " -p <port>          Server port (default 6379)\n"
 " -s <socket>        Server socket (overrides host and port)\n"
 " -a <password>      Password for Redis Auth\n"
 " --user <username>  Used to send ACL style 'AUTH username pass'. Needs -a.\n"
+" -u <uri>           Server URI.\n"
 " -c <clients>       Number of parallel connections (default 50)\n"
 " -n <requests>      Total number of requests (default 100000)\n"
 " -d <size>          Data size of SET/GET value in bytes (default 3)\n"
 " --dbnum <db>       SELECT the specified db number (default 0)\n"
+" -3                 Start session in RESP3 protocol mode.\n"
 " --threads <num>    Enable multi-thread mode.\n"
 " --cluster          Enable cluster mode.\n"
+"                    If the command is supplied on the command line in cluster\n"
+"                    mode, the key must contain \"{tag}\". Otherwise, the\n"
+"                    command will not be sent to the right cluster node.\n"
 " --enable-tracking  Send CLIENT TRACKING on before starting benchmark.\n"
 " -k <boolean>       1=keep alive 0=reconnect (default 1)\n"
 " -r <keyspacelen>   Use random keys for SET/GET/INCR, random values for SADD,\n"
 "                    random members and scores for ZADD.\n"
-"  Using this option the benchmark will expand the string __rand_int__\n"
-"  inside an argument with a 12 digits number in the specified range\n"
-"  from 0 to keyspacelen-1. The substitution changes every time a command\n"
-"  is executed. Default tests use this to hit random keys in the\n"
-"  specified range.\n"
+"                    Using this option the benchmark will expand the string\n"
+"                    __rand_int__ inside an argument with a 12 digits number in\n"
+"                    the specified range from 0 to keyspacelen-1. The\n"
+"                    substitution changes every time a command is executed.\n"
+"                    Default tests use this to hit random keys in the specified\n"
+"                    range.\n"
+"                    Note: If -r is omitted, all commands in a benchmark will\n"
+"                    use the same key.\n"
 " -P <numreq>        Pipeline <numreq> requests. Default 1 (no pipeline).\n"
-" -e                 If server replies with errors, show them on stdout.\n"
-"                    (no more than 1 error per second is displayed)\n"
 " -q                 Quiet. Just show query/sec values\n"
 " --precision        Number of decimal places to display in latency output (default 0)\n"
 " --csv              Output in CSV format\n"
 " -l                 Loop. Run the tests forever\n"
 " -t <tests>         Only run the comma separated list of tests. The test\n"
 "                    names are the same as the ones produced as output.\n"
+"                    The -t option is ignored if a specific command is supplied\n"
+"                    on the command line.\n"
 " -I                 Idle mode. Just open N idle connections and wait.\n"
+" -x                 Read last argument from STDIN.\n"
 #ifdef USE_OPENSSL
 " --tls              Establish a secure TLS connection.\n"
 " --sni <host>       Server name indication for TLS.\n"
@@ -1593,18 +1627,18 @@ usage:
 " --insecure         Allow insecure TLS connection by skipping cert validation.\n"
 " --cert <file>      Client certificate to authenticate with.\n"
 " --key <file>       Private key file to authenticate with.\n"
-" --tls-ciphers <list> Sets the list of prefered ciphers (TLSv1.2 and below)\n"
+" --tls-ciphers <list> Sets the list of preferred ciphers (TLSv1.2 and below)\n"
 "                    in order of preference from highest to lowest separated by colon (\":\").\n"
 "                    See the ciphers(1ssl) manpage for more information about the syntax of this string.\n"
 #ifdef TLS1_3_VERSION
-" --tls-ciphersuites <list> Sets the list of prefered ciphersuites (TLSv1.3)\n"
+" --tls-ciphersuites <list> Sets the list of preferred ciphersuites (TLSv1.3)\n"
 "                    in order of preference from highest to lowest separated by colon (\":\").\n"
 "                    See the ciphers(1ssl) manpage for more information about the syntax of this string,\n"
 "                    and specifically for TLSv1.3 ciphersuites.\n"
 #endif
 #endif
 " --help             Output this help and exit.\n"
-" --version          Output version and exit.\n\n"
+" --version          Output version and exit.\n\n",
 "Examples:\n\n"
 " Run the benchmark with the default configuration against 127.0.0.1:6379:\n"
 "   $ redis-benchmark\n\n"
@@ -1627,7 +1661,7 @@ usage:
 int showThroughput(struct aeEventLoop *eventLoop, long long id, void *clientData) {
     UNUSED(eventLoop);
     UNUSED(id);
-    UNUSED(clientData);
+    benchmarkThread *thread = (benchmarkThread *)clientData;
     int liveclients = 0;
     int requests_finished = 0;
     int previous_requests_finished = 0;
@@ -1635,7 +1669,7 @@ int showThroughput(struct aeEventLoop *eventLoop, long long id, void *clientData
     atomicGet(config.liveclients, liveclients);
     atomicGet(config.requests_finished, requests_finished);
     atomicGet(config.previous_requests_finished, previous_requests_finished);
-    
+
     if (liveclients == 0 && requests_finished != config.requests) {
         fprintf(stderr,"All clients disconnected... aborting.\n");
         exit(1);
@@ -1644,11 +1678,15 @@ int showThroughput(struct aeEventLoop *eventLoop, long long id, void *clientData
         aeStop(eventLoop);
         return AE_NOMORE;
     }
-    if (config.csv) return 250;
+    if (config.csv) return SHOW_THROUGHPUT_INTERVAL;
+    /* only first thread output throughput */
+    if (thread != NULL && thread->index != 0) {
+        return SHOW_THROUGHPUT_INTERVAL;
+    }
     if (config.idlemode == 1) {
         printf("clients: %d\r", config.liveclients);
         fflush(stdout);
-	return 250;
+        return SHOW_THROUGHPUT_INTERVAL;
     }
     const float dt = (float)(current_tick-config.start)/1000.0;
     const float rps = (float)requests_finished/dt;
@@ -1656,15 +1694,17 @@ int showThroughput(struct aeEventLoop *eventLoop, long long id, void *clientData
     const float instantaneous_rps = (float)(requests_finished-previous_requests_finished)/instantaneous_dt;
     config.previous_tick = current_tick;
     atomicSet(config.previous_requests_finished,requests_finished);
-    config.last_printed_bytes = printf("%s: rps=%.1f (overall: %.1f) avg_msec=%.3f (overall: %.3f)\r", config.title, instantaneous_rps, rps, hdr_mean(config.current_sec_latency_histogram)/1000.0f, hdr_mean(config.latency_histogram)/1000.0f);
+    printf("%*s\r", config.last_printed_bytes, " "); /* ensure there is a clean line */
+    int printed_bytes = printf("%s: rps=%.1f (overall: %.1f) avg_msec=%.3f (overall: %.3f)\r", config.title, instantaneous_rps, rps, hdr_mean(config.current_sec_latency_histogram)/1000.0f, hdr_mean(config.latency_histogram)/1000.0f);
+    config.last_printed_bytes = printed_bytes;
     hdr_reset(config.current_sec_latency_histogram);
     fflush(stdout);
-    return 250; /* every 250ms */
+    return SHOW_THROUGHPUT_INTERVAL;
 }
 
 /* Return true if the named test was selected using the -t command line
  * switch, or if all the tests are selected (no -t passed by user). */
-int test_is_selected(char *name) {
+int test_is_selected(const char *name) {
     char buf[256];
     int l = strlen(name);
 
@@ -1676,7 +1716,7 @@ int test_is_selected(char *name) {
     return strstr(config.tests,buf) != NULL;
 }
 
-int main(int argc, const char **argv) {
+int main(int argc, char **argv) {
     int i;
     char *data, *cmd, *tag;
     int len;
@@ -1697,7 +1737,6 @@ int main(int argc, const char **argv) {
     config.keepalive = 1;
     config.datasize = 3;
     config.pipeline = 1;
-    config.showerrors = 0;
     config.randomkeys = 0;
     config.randomkeys_keyspacelen = 0;
     config.quiet = 0;
@@ -1705,12 +1744,13 @@ int main(int argc, const char **argv) {
     config.loop = 0;
     config.idlemode = 0;
     config.clients = listCreate();
-    config.hostip = "127.0.0.1";
-    config.hostport = 6379;
+    config.conn_info.hostip = sdsnew("127.0.0.1");
+    config.conn_info.hostport = 6379;
     config.hostsocket = NULL;
     config.tests = NULL;
-    config.dbnum = 0;
-    config.auth = NULL;
+    config.conn_info.input_dbnum = 0;
+    config.stdinarg = 0;
+    config.conn_info.auth = NULL;
     config.precision = DEFAULT_LATENCY_PRECISION;
     config.num_threads = 0;
     config.threads = NULL;
@@ -1722,6 +1762,7 @@ int main(int argc, const char **argv) {
     config.is_updating_slots = 0;
     config.slots_last_update = 0;
     config.enable_tracking = 0;
+    config.resp3 = 0;
 
     i = parseOptions(argc,argv);
     argc -= i;
@@ -1743,7 +1784,7 @@ int main(int argc, const char **argv) {
         if (!fetchClusterConfiguration() || !config.cluster_nodes) {
             if (!config.hostsocket) {
                 fprintf(stderr, "Failed to fetch cluster configuration from "
-                                "%s:%d\n", config.hostip, config.hostport);
+                                "%s:%d\n", config.conn_info.hostip, config.conn_info.hostport);
             } else {
                 fprintf(stderr, "Failed to fetch cluster configuration from "
                                 "%s\n", config.hostsocket);
@@ -1768,7 +1809,7 @@ int main(int argc, const char **argv) {
             printf("%s:%d\n", node->ip, node->port);
             node->redis_config = getRedisConfig(node->ip, node->port, NULL);
             if (node->redis_config == NULL) {
-                fprintf(stderr, "WARN: could not fetch node CONFIG %s:%d\n",
+                fprintf(stderr, "WARNING: Could not fetch node CONFIG %s:%d\n",
                         node->ip, node->port);
             }
         }
@@ -1779,9 +1820,10 @@ int main(int argc, const char **argv) {
             config.num_threads = config.cluster_node_count;
     } else {
         config.redis_config =
-            getRedisConfig(config.hostip, config.hostport, config.hostsocket);
-        if (config.redis_config == NULL)
-            fprintf(stderr, "WARN: could not fetch server CONFIG\n");
+            getRedisConfig(config.conn_info.hostip, config.conn_info.hostport, config.hostsocket);
+        if (config.redis_config == NULL) {
+            fprintf(stderr, "WARNING: Could not fetch server CONFIG\n");
+        }
     }
     if (config.num_threads > 0) {
         pthread_mutex_init(&(config.liveclients_mutex), NULL);
@@ -1789,7 +1831,14 @@ int main(int argc, const char **argv) {
     }
 
     if (config.keepalive == 0) {
-        printf("WARNING: keepalive disabled, you probably need 'echo 1 > /proc/sys/net/ipv4/tcp_tw_reuse' for Linux and 'sudo sysctl -w net.inet.tcp.msl=1000' for Mac OS X in order to use a lot of clients/requests\n");
+        fprintf(stderr,
+                "WARNING: Keepalive disabled. You probably need "
+                "'echo 1 > /proc/sys/net/ipv4/tcp_tw_reuse' for Linux and "
+                "'sudo sysctl -w net.inet.tcp.msl=1000' for Mac OS X in order "
+                "to use a lot of clients/requests\n");
+    }
+    if (argc > 0 && config.tests != NULL) {
+        fprintf(stderr, "WARNING: Option -t is ignored.\n");
     }
 
     if (config.idlemode) {
@@ -1815,15 +1864,26 @@ int main(int argc, const char **argv) {
             title = sdscatlen(title, " ", 1);
             title = sdscatlen(title, (char*)argv[i], strlen(argv[i]));
         }
-
+        sds *sds_args = getSdsArrayFromArgv(argc, argv, 0);
+        if (!sds_args) {
+            fprintf(stderr, "Invalid quoted string\n");
+            return 1;
+        }
+        if (config.stdinarg) {
+            sds_args = sds_realloc(sds_args,(argc + 1) * sizeof(sds));
+            sds_args[argc] = readArgFromStdin();
+            argc++;
+        }
         do {
-            len = redisFormatCommandArgv(&cmd,argc,argv,NULL);
+            len = redisFormatCommandArgv(&cmd,argc,(const char**)sds_args,NULL);
             // adjust the datasize to the parsed command
             config.datasize = len;
             benchmark(title,cmd,len);
             free(cmd);
         } while(config.loop);
+        sdsfreesplitres(sds_args, argc);
 
+        sdsfree(title);
         if (config.redis_config != NULL) freeRedisConfig(config.redis_config);
         return 0;
     }
@@ -1944,8 +2004,8 @@ int main(int argc, const char **argv) {
         }
 
         if (test_is_selected("lrange") || test_is_selected("lrange_500")) {
-            len = redisFormatCommand(&cmd,"LRANGE mylist%s 0 449",tag);
-            benchmark("LRANGE_500 (first 450 elements)",cmd,len);
+            len = redisFormatCommand(&cmd,"LRANGE mylist%s 0 499",tag);
+            benchmark("LRANGE_500 (first 500 elements)",cmd,len);
             free(cmd);
         }
 
@@ -1972,6 +2032,8 @@ int main(int argc, const char **argv) {
         if (!config.csv) printf("\n");
     } while(config.loop);
 
+    zfree(data);
+    freeCliConnInfo(config.conn_info);
     if (config.redis_config != NULL) freeRedisConfig(config.redis_config);
 
     return 0;
