@@ -40,20 +40,20 @@ start_server {tags {"tracking network"}} {
     } {*OK}
 
     test {The other connection is able to get invalidations} {
-        r SET a 1
-        r SET b 1
-        r GET a
-        r INCR b ; # This key should not be notified, since it wasn't fetched.
-        r INCR a
+        r SET a{t} 1
+        r SET b{t} 1
+        r GET a{t}
+        r INCR b{t} ; # This key should not be notified, since it wasn't fetched.
+        r INCR a{t}
         set keys [lindex [$rd_redirection read] 2]
         assert {[llength $keys] == 1}
-        assert {[lindex $keys 0] eq {a}}
+        assert {[lindex $keys 0] eq {a{t}}}
     }
 
     test {The client is now able to disable tracking} {
         # Make sure to add a few more keys in the tracking list
         # so that we can check for leaks, as a side effect.
-        r MGET a b c d e f g
+        r MGET a{t} b{t} c{t} d{t} e{t} f{t} g{t}
         r CLIENT TRACKING off
     } {*OK}
 
@@ -62,28 +62,28 @@ start_server {tags {"tracking network"}} {
     } {*OK*}
 
     test {The connection gets invalidation messages about all the keys} {
-        r MSET a 1 b 2 c 3
+        r MSET a{t} 1 b{t} 2 c{t} 3
         set keys [lsort [lindex [$rd_redirection read] 2]]
-        assert {$keys eq {a b c}}
+        assert {$keys eq {a{t} b{t} c{t}}}
     }
 
     test {Clients can enable the BCAST mode with prefixes} {
         r CLIENT TRACKING off
         r CLIENT TRACKING on BCAST REDIRECT $redir_id PREFIX a: PREFIX b:
         r MULTI
-        r INCR a:1
-        r INCR a:2
-        r INCR b:1
-        r INCR b:2
+        r INCR a:1{t}
+        r INCR a:2{t}
+        r INCR b:1{t}
+        r INCR b:2{t}
         # we should not get this key
-        r INCR c:1
+        r INCR c:1{t}
         r EXEC
         # Because of the internals, we know we are going to receive
         # two separated notifications for the two different prefixes.
         set keys1 [lsort [lindex [$rd_redirection read] 2]]
         set keys2 [lsort [lindex [$rd_redirection read] 2]]
         set keys [lsort [list {*}$keys1 {*}$keys2]]
-        assert {$keys eq {a:1 a:2 b:1 b:2}}
+        assert {$keys eq {a:1{t} a:2{t} b:1{t} b:2{t}}}
     }
 
     test {Adding prefixes to BCAST mode works} {
@@ -96,16 +96,16 @@ start_server {tags {"tracking network"}} {
     test {Tracking NOLOOP mode in standard mode works} {
         r CLIENT TRACKING off
         r CLIENT TRACKING on REDIRECT $redir_id NOLOOP
-        r MGET otherkey1 loopkey otherkey2
-        $rd_sg SET otherkey1 1; # We should get this
-        r SET loopkey 1 ; # We should not get this
-        $rd_sg SET otherkey2 1; # We should get this
+        r MGET otherkey1{t} loopkey{t} otherkey2{t}
+        $rd_sg SET otherkey1{t} 1; # We should get this
+        r SET loopkey{t} 1 ; # We should not get this
+        $rd_sg SET otherkey2{t} 1; # We should get this
         # Because of the internals, we know we are going to receive
         # two separated notifications for the two different keys.
         set keys1 [lsort [lindex [$rd_redirection read] 2]]
         set keys2 [lsort [lindex [$rd_redirection read] 2]]
         set keys [lsort [list {*}$keys1 {*}$keys2]]
-        assert {$keys eq {otherkey1 otherkey2}}
+        assert {$keys eq {otherkey1{t} otherkey2{t}}}
     }
 
     test {Tracking NOLOOP mode in BCAST mode works} {
@@ -208,17 +208,31 @@ start_server {tags {"tracking network"}} {
         assert {$res eq {key1}}
     }
 
-    test {Invalid keys should not be tracked for scripts in NOLOOP mode} {
-        $rd_sg CLIENT TRACKING off
-        $rd_sg CLIENT TRACKING on NOLOOP
-        $rd_sg HELLO 3
-        $rd_sg SET key1 1
-        assert_equal "1" [$rd_sg GET key1]
+    test {Tracking only occurs for scripts when a command calls a read-only command} {
+        r CLIENT TRACKING off
+        r CLIENT TRACKING on
+        $rd_sg MSET key2{t} 1 key2{t} 1
 
-        # For write command in script, invalid key should not be tracked with NOLOOP flag
-        $rd_sg eval "return redis.call('set', 'key1', '2')" 1 key1
-        assert_equal "2" [$rd_sg GET key1]
-        $rd_sg CLIENT TRACKING off
+        # If a script doesn't call any read command, don't track any keys
+        r EVAL "redis.call('set', 'key3{t}', 'bar')" 2 key1{t} key2{t} 
+        $rd_sg MSET key2{t} 2 key1{t} 2
+
+        # If a script calls a read command, track all declared keys
+        r EVAL "redis.call('get', 'key3{t}')" 2 key1{t} key2{t} 
+        $rd_sg MSET key2{t} 2 key1{t} 2
+        assert_equal {invalidate key2{t}} [r read]
+        assert_equal {invalidate key1{t}} [r read]
+
+        # RO variants work like the normal variants
+        r EVAL_RO "redis.call('ping')" 2 key1{t} key2{t} 
+        $rd_sg MSET key2{t} 2 key1{t} 2
+
+        r EVAL_RO "redis.call('get', 'key1{t}')" 2 key1{t} key2{t} 
+        $rd_sg MSET key2{t} 3 key1{t} 3
+        assert_equal {invalidate key2{t}} [r read]
+        assert_equal {invalidate key1{t}} [r read]
+
+        assert_equal "PONG" [r ping]
     }
 
     test {RESP3 Client gets tracking-redir-broken push message after cached key changed when rediretion client is terminated} {
@@ -249,16 +263,16 @@ start_server {tags {"tracking network"}} {
         r CLIENT TRACKING on REDIRECT $redir_id
         $rd CLIENT TRACKING on REDIRECT $redir_id 
         assert_equal OK [$rd read] ; # Consume the TRACKING reply
-        $rd_sg MSET key1 1 key2 1
-        r GET key1
-        $rd GET key2 
+        $rd_sg MSET key1{t} 1 key2{t} 1
+        r GET key1{t}
+        $rd GET key2{t} 
         assert_equal 1 [$rd read] ; # Consume the GET reply
-        $rd_sg INCR key1
-        $rd_sg INCR key2
+        $rd_sg INCR key1{t}
+        $rd_sg INCR key2{t}
         set res1 [lindex [$rd_redirection read] 2]
         set res2 [lindex [$rd_redirection read] 2]
-        assert {$res1 eq {key1}}
-        assert {$res2 eq {key2}}
+        assert {$res1 eq {key1{t}}}
+        assert {$res2 eq {key2{t}}}
     }
 
     test {Different clients using different protocols can track the same key} {
@@ -382,7 +396,7 @@ start_server {tags {"tracking network"}} {
         $r CLIENT TRACKING OFF
     }
 
-    test {hdel deliver invlidate message after response in the same connection} {
+    test {hdel deliver invalidate message after response in the same connection} {
         r CLIENT TRACKING off
         r HELLO 3
         r CLIENT TRACKING on
@@ -437,7 +451,11 @@ start_server {tags {"tracking network"}} {
         r CLIENT TRACKING off
         r HELLO 3
         r CLIENT TRACKING on
-        set used [s used_memory]
+
+        # make the previous test is really done before sampling used_memory
+        wait_lazyfree_done r
+
+        set used [expr {[s used_memory] - [s mem_not_counted_for_evict]}]
         set limit [expr {$used+100*1024}]
         set old_policy [lindex [r config get maxmemory-policy] 1]
         r config set maxmemory $limit
@@ -449,7 +467,7 @@ start_server {tags {"tracking network"}} {
         # We use SETBIT here, so we can set a big key and get the used_memory
         # bigger than maxmemory. Next command will evict volatile keys. We
         # can't use SET, as SET uses big input buffer, so it will fail.
-        r setbit big-key 1000000 0
+        r setbit big-key 1600000 0 ;# this will consume 200kb
         # volatile-key is evicted before response.
         set res [r getbit big-key 0]
         assert_equal $res {invalidate volatile-key}
@@ -480,9 +498,9 @@ start_server {tags {"tracking network"}} {
     test {Tracking gets notification on tracking table key eviction} {
         r CLIENT TRACKING off
         r CLIENT TRACKING on REDIRECT $redir_id NOLOOP
-        r MSET key1 1 key2 2
+        r MSET key1{t} 1 key2{t} 2
         # Let the server track the two keys for us
-        r MGET key1 key2
+        r MGET key1{t} key2{t}
         # Force the eviction of all the keys but one:
         r config set tracking-table-max-keys 1
         # Note that we may have other keys in the table for this client,
@@ -492,11 +510,11 @@ start_server {tags {"tracking network"}} {
         # otherwise the test will die for timeout.
         while 1 {
             set keys [lindex [$rd_redirection read] 2]
-            if {$keys eq {key1} || $keys eq {key2}} break
+            if {$keys eq {key1{t}} || $keys eq {key2{t}}} break
         }
         # We should receive an expire notification for one of
         # the two keys (only one must remain)
-        assert {$keys eq {key1} || $keys eq {key2}}
+        assert {$keys eq {key1{t}} || $keys eq {key2{t}}}
     }
 
     test {Invalidation message received for flushall} {
@@ -530,20 +548,6 @@ start_server {tags {"tracking network"}} {
         assert_equal [lindex [$rd_redirection read] 2] {}
     }
 
-    test {flushdb tracking invalidation message is not interleaved with transaction response} {
-        clean_all
-        r HELLO 3
-        r CLIENT TRACKING on
-        r SET a{t} 1
-        r GET a{t}
-        r MULTI
-        r FLUSHDB
-        set res [r EXEC]
-        assert_equal $res {OK}
-        # Consume the invalidate message which is after command response
-        r read
-    } {invalidate {}}
-
     # Keys are defined to be evicted 100 at a time by default.
     # If after eviction the number of keys still surpasses the limit
     # defined in tracking-table-max-keys, we increases eviction 
@@ -708,135 +712,6 @@ start_server {tags {"tracking network"}} {
         assert_equal {} $prefixes
     }
 
-    test {Regression test for #11715} {
-        # This issue manifests when a client invalidates keys through the max key
-        # limit, which invalidates keys to get Redis below the limit, but no command is
-        # then executed. This can occur in several ways but the simplest is through 
-        # multi-exec which queues commands.
-        clean_all
-        r config set tracking-table-max-keys 2
-
-        # The cron will invalidate keys if we're above the limit, so disable it.
-        r debug pause-cron 1
-
-        # Set up a client that has listened to 2 keys and start a multi, this
-        # sets up the crash for later.
-        $rd HELLO 3
-        $rd read
-        $rd CLIENT TRACKING on
-        assert_match "OK" [$rd read]
-        $rd mget "1{tag}" "2{tag}"
-        assert_match "{} {}" [$rd read]
-        $rd multi
-        assert_match "OK" [$rd read]
-
-        # Reduce the tracking table keys to 1, this doesn't immediately take affect, but
-        # instead will apply on the next command.
-        r config set tracking-table-max-keys 1
-
-        # This command will get queued, so make sure this command doesn't crash.
-        $rd ping
-        $rd exec
-
-        # Validate we got some invalidation message and then the command was queued.
-        assert_match "invalidate *{tag}" [$rd read]
-        assert_match "QUEUED" [$rd read]
-        assert_match "PONG" [$rd read]
-
-        r debug pause-cron 0
-    } {OK} {needs:debug}
-
-    foreach resp {3 2} {
-        test "RESP$resp based basic invalidation with client reply off" {
-            # This entire test is mostly irrelevant for RESP2, but we run it anyway just for some extra coverage.
-            clean_all
-
-            $rd hello $resp
-            $rd read
-            $rd client tracking on
-            $rd read
-
-            $rd_sg set foo bar
-            $rd get foo
-            $rd read
-
-            $rd client reply off
-
-            $rd_sg set foo bar2
-
-            if {$resp == 3} {
-                assert_equal {invalidate foo} [$rd read]
-            } elseif {$resp == 2} { } ;# Just coverage
-
-            # Verify things didn't get messed up and no unexpected reply was pushed to the client.
-            $rd client reply on
-            assert_equal {OK} [$rd read]
-            $rd ping
-            assert_equal {PONG} [$rd read]
-        }
-    }
-
-    test {RESP3 based basic redirect invalidation with client reply off} {
-        clean_all
-
-        set rd_redir [redis_deferring_client]
-        $rd_redir hello 3
-        $rd_redir read
-
-        $rd_redir client id
-        set rd_redir_id [$rd_redir read]
-
-        $rd client tracking on redirect $rd_redir_id
-        $rd read
-
-        $rd_sg set foo bar
-        $rd get foo
-        $rd read
-
-        $rd_redir client reply off
-
-        $rd_sg set foo bar2
-        assert_equal {invalidate foo} [$rd_redir read]
-
-        # Verify things didn't get messed up and no unexpected reply was pushed to the client.
-        $rd_redir client reply on
-        assert_equal {OK} [$rd_redir read]
-        $rd_redir ping
-        assert_equal {PONG} [$rd_redir read]
-
-        $rd_redir close
-    }
-
-    test {RESP3 based basic tracking-redir-broken with client reply off} {
-        clean_all
-
-        $rd hello 3
-        $rd read
-        $rd client tracking on redirect $redir_id
-        $rd read
-
-        $rd_sg set foo bar
-        $rd get foo
-        $rd read
-
-        $rd client reply off
-
-        $rd_redirection quit
-        $rd_redirection read
-
-        $rd_sg set foo bar2
-
-        set res [lsearch -exact [$rd read] "tracking-redir-broken"]
-        assert_morethan_equal $res 0
-
-        # Verify things didn't get messed up and no unexpected reply was pushed to the client.
-        $rd client reply on
-        assert_equal {OK} [$rd read]
-        $rd ping
-        assert_equal {PONG} [$rd read]
-    }
-
     $rd_redirection close
-    $rd_sg close
     $rd close
 }
